package sdanderson.androidcharts.LineChart;

import android.content.Context;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.RectF;
import android.graphics.Region;
import android.util.AttributeSet;
import android.view.View;

import java.util.ArrayList;
import java.util.List;

import sdanderson.androidcharts.Common.Coordinate;
import sdanderson.androidcharts.Common.CoreSettings;
import sdanderson.androidcharts.Common.NumberUtilities;

/**
 * Created by Shane Anderson on 9/30/2015.
 *
 * This class is to provide the view into the Line Chart.
 *
 */
class LineChartInnerView extends View {

    private float mBorderWidth = 6f;
    private int mBorderColor = Color.BLACK;
    private int mBackgroundColor = Color.WHITE;
    private int mGridlineColor = Color.GRAY;
    private float mGridlineWidth = 1f;
    private int mNumOfGridlines = 5;
    private int mZeroLineColor = Color.BLACK;
    private float mZeroLineWidth = 4f;
    private boolean mUseAnimation = false;
    private float mPadding = mBorderWidth / 2f;

    private RectF mRectF = new RectF();

    private List<List<Region>> mTouchPoints;
    private List<LineChartItem> mLines;

    private Paint mDrawingPaint = new Paint();
    private RectF mDrawingRect = new RectF();

    private boolean mRoundOutChartYMinMax = false;
    private int mRoundOutChartYMinMaxHighestDigits = 1;
    private boolean mRoundOutChartXMinMax = false;
    private int mRoundOutChartXMinMaxHighestDigits = 1;
    private float mMinimumXValue = 0f;
    private float mMaximumXValue = 0f;
    private float mMinimumYValue = 0f;
    private float mMaximumYValue = 0f;
    private float mXScaling = 1f;
    private float mYScaling = 1f;

    /* Variables to support animations. The animation sweeps along the X axis from left to right,
     * drawing all lines from mMinimumXValue to mSweepX. mSweepXIncrement is how many pixels the
     * sweep operation should move across the chart on each draw. mSweepXDamper is inversely
     * proportional to mSweepXIncrement; the higher mSweepXDamper, the slower the animation
     * will proceed.
     */
    private boolean mFinishedDrawing = false;
    private float mAnimationTime = 500f; // in milliseconds
    private long mAnimationElapsedTime = 0;
    private long mAnimationStartTime = 0;
    private boolean mAnimationCounterStarted = false;
    private float mSweepX = 0f;
    private boolean mFirstDrawPass = true;
    Coordinate<Float, Float> mAnimCoord1 = new Coordinate<>(0f, 0f);
    Coordinate<Float, Float> mAnimCoord2 = new Coordinate<>(0f, 0f);
    private Runnable mRedrawRunnable = new Runnable() {
        public void run() {
            try {
                if (!mFinishedDrawing) {
                    if (!mAnimationCounterStarted) {
                        mAnimationStartTime = System.currentTimeMillis();
                        mAnimationCounterStarted = true;
                    }
                    mAnimationElapsedTime = System.currentTimeMillis() - mAnimationStartTime;
                    invalidate();
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    };

    public LineChartInnerView(Context pContext) {
        super(pContext);
        init();
    }

    public LineChartInnerView(Context pContext, AttributeSet pAttrs) {
        super(pContext, pAttrs);
        init();
    }

    public LineChartInnerView(Context pContext, AttributeSet pAttrs, int pDefStyle) {
        super(pContext, pAttrs, pDefStyle);
        init();
    }

    private void init() {
        mLines = new ArrayList<>();
        setWillNotDraw(false);
        this.setBackgroundColor(Color.TRANSPARENT);
    }

    @Override
    public void onDraw(final Canvas pCanvas) {
        if (pCanvas == null)
            return;

        mPadding = mBorderWidth / 2f;
        pCanvas.drawColor(mBackgroundColor);

        if (mRectF != null && mDrawingPaint != null && !this.isInEditMode()) {
            mRectF.set(mPadding, mPadding, getWidth() - mPadding, getHeight() - mPadding);

            // The Titles and Axis text is handled by the LineChart ViewGroup

            calculateScaling();

            // Draw the horizontal grid lines and the Y=0 line if necessary

                // Draw a line at Y=0 if there is a mix of negative and positive values
            if (mMaximumYValue > 0f && mMinimumYValue < 0f) {
                mDrawingPaint.setColor(mZeroLineColor);
                mDrawingPaint.setStrokeWidth(mZeroLineWidth);
                mDrawingPaint.setStyle(Paint.Style.STROKE);
                mDrawingPaint.setAntiAlias(true);
                pCanvas.drawLine(0f, actualYToDrawingY(0f), mRectF.width(), actualYToDrawingY(0f),
                                                                mDrawingPaint);

                // Draw horizontal grid lines on both side of Y=0
                // First draw horizontal grid lines for above Y=0
                Float spacing = (actualYToDrawingY(0f)-mRectF.top) / (mNumOfGridlines+1);
                for (int i = 1; i <= mNumOfGridlines; i++) {
                    mDrawingPaint.setColor(mGridlineColor);
                    mDrawingPaint.setStrokeWidth(mGridlineWidth);
                    mDrawingPaint.setStyle(Paint.Style.STROKE);
                    mDrawingPaint.setAntiAlias(true);
                    pCanvas.drawLine(mRectF.left, mRectF.top+(spacing*i), mRectF.right,
                            mRectF.top+(spacing*i), mDrawingPaint);
                }

                // Now draw horizontal grid lines for below Y=0
                spacing = (mRectF.bottom-actualYToDrawingY(0f)) / (mNumOfGridlines+1);
                for (int i = 1; i <= mNumOfGridlines; i++) {
                    mDrawingPaint.setColor(mGridlineColor);
                    mDrawingPaint.setStrokeWidth(mGridlineWidth);
                    mDrawingPaint.setStyle(Paint.Style.STROKE);
                    mDrawingPaint.setAntiAlias(true);
                    pCanvas.drawLine(mRectF.left, mRectF.bottom-(spacing*i), mRectF.right,
                            mRectF.bottom-(spacing*i), mDrawingPaint);
                }
            } else {
                // In this case all values are either positive or all are negative, so just space
                // gridlines out evenly from the max to min
                float spacing = mRectF.height() / (mNumOfGridlines+1);
                for (int i = 1; i <= mNumOfGridlines; i++) {
                    mDrawingPaint.setColor(mGridlineColor);
                    mDrawingPaint.setStrokeWidth(mGridlineWidth);
                    mDrawingPaint.setStyle(Paint.Style.STROKE);
                    mDrawingPaint.setAntiAlias(true);
                    pCanvas.drawLine(mRectF.left, mRectF.top + (spacing * i), mRectF.right,
                            mRectF.top + (spacing * i), mDrawingPaint);
                }
            }

            if (!mUseAnimation) {

                // Draw the lines without any animations
                if (mLines != null && !mLines.isEmpty()) {
                    for (int i = 0; i < mLines.size(); i++) {

                        // Skip any data sets with only one entry. Can't chart that!
                        if (mLines.get(i).getCoordinates().size() > 1) {
                            mDrawingPaint.setColor(mLines.get(i).getLineColor());
                            mDrawingPaint.setStrokeWidth(mLines.get(i).getLineThickness());
                            mDrawingPaint.setStyle(Paint.Style.STROKE);
                            mDrawingPaint.setAntiAlias(true);

                            for (int j = 1; j < mLines.get(i).getCoordinates().size(); j++) {

                                // Lines
                                Float x1 = actualXToDrawingX(
                                        mLines.get(i).getCoordinatesAtIndex(j - 1).getX());
                                Float x2 = actualXToDrawingX(
                                        mLines.get(i).getCoordinatesAtIndex(j).getX());
                                Float y1 = actualYToDrawingY(
                                        mLines.get(i).getCoordinatesAtIndex(j - 1).getY());
                                Float y2 = actualYToDrawingY(
                                        mLines.get(i).getCoordinatesAtIndex(j).getY());

                                pCanvas.drawLine(x1, y1, x2, y2, mDrawingPaint);

                                // Draw a dot at each segment with a touch point.
                                // The touch point is a 2d array: first indexes which LineChartItem
                                // then which coordinate within the LineChartItem it corresponds to.
                                mDrawingPaint.setStyle(Paint.Style.FILL);
                                Float radius = mLines.get(i).getLineThickness() * 3f;
                                pCanvas.drawCircle(x1, y1, radius, mDrawingPaint);
                                mTouchPoints.get(i).get(j - 1).set((int) (x1 - radius * 2),
                                        (int) (y1 - radius * 2),
                                        (int) (x1 + radius * 2),
                                        (int) (y1 + radius * 2));

                                if (j == mLines.get(i).getCoordinates().size() - 1) {
                                    pCanvas.drawCircle(x2, y2, radius, mDrawingPaint);
                                    mTouchPoints.get(i).get(j).set((int) (x2 - radius * 2),
                                            (int) (y2 - radius * 2),
                                            (int) (x2 + radius * 2),
                                            (int) (y2 + radius * 2));
                                }
                            }
                        }
                    }
                }
            } else {
                // Draw the lines using animations
                if (mLines != null && !mLines.isEmpty()) {
                    for (int i = 0; i < mLines.size(); i++) {

                        // Skip any data sets with only one entry. Can't chart that!
                        if (mLines.get(i).getCoordinates().size() > 1) {
                            mDrawingPaint.setColor(mLines.get(i).getLineColor());
                            mDrawingPaint.setStrokeWidth(mLines.get(i).getLineThickness());
                            mDrawingPaint.setStyle(Paint.Style.STROKE);
                            mDrawingPaint.setAntiAlias(true);

                            calculateScaling();

                            if (mFirstDrawPass) {
                                mSweepX = actualXToDrawingX(mMinimumXValue);
                            }

                            for (int j = 1; j < mLines.get(i).getCoordinates().size(); j++) {

                                // Lines
                                Float x1 = actualXToDrawingX(
                                        mLines.get(i).getCoordinatesAtIndex(j - 1).getX());
                                Float y1 = actualYToDrawingY(
                                        mLines.get(i).getCoordinatesAtIndex(j - 1).getY());
                                Float x2 = actualXToDrawingX(
                                        mLines.get(i).getCoordinatesAtIndex(j).getX());
                                Float y2 = actualYToDrawingY(
                                        mLines.get(i).getCoordinatesAtIndex(j).getY());
                                if (mSweepX > x1 && mSweepX < x2) {
                                    mAnimCoord1.setX(x1);
                                    mAnimCoord1.setY(y1);
                                    mAnimCoord2.setX(x2);
                                    mAnimCoord2.setY(y2);
                                    x2 = mSweepX;
                                    y2 = interpolateYFromX(mSweepX, mAnimCoord1, mAnimCoord2);
                                }

                                // only bother drawing a line if the x is prior to our SweepX
                                if (x1 < mSweepX) {
                                    pCanvas.drawLine(x1, y1, x2, y2, mDrawingPaint);
                                }
                            }
                        }

                    }
                }

                if (mLines != null && !mLines.isEmpty()) {
                    for (int i = 0; i < mLines.size(); i++) {
                        for (int j = 1; j < mLines.get(i).getCoordinates().size(); j++) {

                            Float x1 = actualXToDrawingX(
                                    mLines.get(i).getCoordinatesAtIndex(j - 1).getX());
                            Float x2 = actualXToDrawingX(
                                    mLines.get(i).getCoordinatesAtIndex(j).getX());
                            Float y1 = actualYToDrawingY(
                                    mLines.get(i).getCoordinatesAtIndex(j - 1).getY());
                            Float y2 = actualYToDrawingY(
                                    mLines.get(i).getCoordinatesAtIndex(j).getY());

                            // Draw a dot at each segment with a touch point.
                            // mTouchPoints is a 2d array that first indexes the LineChartItem
                            // then which coordinate within the LineChartItem it corresponds to.

                            // Only place circles at points passed by mSweepX
                            if (x1 < mSweepX) {
                                mDrawingPaint.setColor(mLines.get(i).getLineColor());
                                mDrawingPaint.setStrokeWidth(mLines.get(i).getLineThickness());
                                mDrawingPaint.setStyle(Paint.Style.FILL);
                                Float radius = mLines.get(i).getLineThickness() * 3f;
                                pCanvas.drawCircle(x1, y1, radius, mDrawingPaint);
                                // Only place the touch points when we've finished animating
                                if (mSweepX == actualXToDrawingX(mMaximumXValue)) {
                                    mTouchPoints.get(i).get(j - 1).set((int) (x1 - radius * 2),
                                            (int) (y1 - radius * 2),
                                            (int) (x1 + radius * 2),
                                            (int) (y1 + radius * 2));
                                }

                                if (j == mLines.get(i).getCoordinates().size() - 1 &&
                                        x2 <= mSweepX) {
                                    pCanvas.drawCircle(x2, y2, radius, mDrawingPaint);
                                    // Only place the touch points when we've finished animating
                                    if (mSweepX == actualXToDrawingX(mMaximumXValue)) {
                                        mTouchPoints.get(i).get(j).set((int) (x2 - radius * 2),
                                                (int) (y2 - radius * 2),
                                                (int) (x2 + radius * 2),
                                                (int) (y2 + radius * 2));
                                    }
                                }
                            }
                        }
                    }
                }

                if (mSweepX == actualXToDrawingX(mMaximumXValue)) {
                    mFinishedDrawing = true;
                }

                // Update the SweepX Location
                if (mSweepX < actualXToDrawingX(mMaximumXValue) &&
                        (actualXToDrawingX(mMaximumXValue)*(mAnimationElapsedTime/mAnimationTime))
                                <= actualXToDrawingX(mMaximumXValue))
                    mSweepX = actualXToDrawingX(mMaximumXValue)*
                                                            (mAnimationElapsedTime/mAnimationTime);
                else {
                    mSweepX = actualXToDrawingX(mMaximumXValue);
                }

                mFirstDrawPass = false;
                postDelayed(mRedrawRunnable, CoreSettings.ANIMATION_RUNNABLE_RESOLUTION);
            }

            // Draw the border
            mDrawingPaint.setColor(mBorderColor);
            mDrawingPaint.setStrokeWidth(mBorderWidth);
            mDrawingPaint.setStyle(Paint.Style.STROKE);
            mDrawingPaint.setAntiAlias(true);
            pCanvas.drawRect(mRectF.left, mRectF.top, mRectF.right, mRectF.bottom, mDrawingPaint);

        } else if (this.isInEditMode()) {
            //The default view which is used, for example, in the layout preview
            mDrawingRect.set(mPadding, mPadding, getWidth() - mPadding, getHeight() - mPadding);
            mDrawingPaint.setColor(Color.BLACK);
            mDrawingPaint.setStrokeWidth(2.0f);
            mDrawingPaint.setStyle(Paint.Style.STROKE);
            mDrawingPaint.setAntiAlias(true);
            pCanvas.drawRect(mDrawingRect.left, mDrawingRect.top, mDrawingRect.right,
                    mDrawingRect.bottom, mDrawingPaint);

            float spacing = getHeight() / 5;
            for (int i = 1; i <= 4; i++) {
                mDrawingPaint.setColor(Color.GRAY);
                mDrawingPaint.setStrokeWidth(1.0f);
                mDrawingPaint.setStyle(Paint.Style.STROKE);
                mDrawingPaint.setAntiAlias(true);
                pCanvas.drawLine(0f, spacing * i, getWidth(), spacing * i, mDrawingPaint);
            }
        }
    }

    protected boolean areThereLines() {
        return (mLines != null && mLines.size() > 0);
    }

    protected void addLine(String pLabel, List<Coordinate<Float, Float>> pCoordinates, int pColor) {
        LineChartItem newLineChartItem = new LineChartItem(pLabel, pCoordinates, pColor);
        mLines.add(newLineChartItem);

        // Check each mValue in the series and update the maximum & minimum X&Y values accordingly.
        // Don't bother adding any lines that have no coordinates.

        if (pCoordinates.size() > 0 && mLines != null) {
            if (mLines.size() == 1) {
                mMinimumYValue = mMaximumYValue = pCoordinates.get(0).getY();
                mMinimumXValue = mMaximumXValue = pCoordinates.get(0).getX();
            }

            // Add the corresponding touch points
            if (mTouchPoints == null)
                mTouchPoints = new ArrayList<>();

            // Add a new ArrayList for the current line in mLines
            mTouchPoints.add(new ArrayList<Region>());

            for (Coordinate<Float, Float> tempCoordinate : pCoordinates) {

                if (tempCoordinate.getY() > mMaximumYValue)
                    mMaximumYValue = tempCoordinate.getY();
                else if (tempCoordinate.getY() < mMinimumYValue)
                    mMinimumYValue = tempCoordinate.getY();

                if (tempCoordinate.getX() > mMaximumXValue)
                    mMaximumXValue = tempCoordinate.getX();
                else if (tempCoordinate.getX() < mMinimumXValue)
                    mMinimumXValue = tempCoordinate.getX();

                // Add each region, the Draw method will calculate the proper sizes
                Region tempRegion = new Region();
                mTouchPoints.get(mTouchPoints.size()-1).add(tempRegion);
            }
        }
    }

    // Here the return mValue is X mValue = LineIndex, Y mValue = Coord Index
    protected List<Coordinate<Integer, Integer>> getTappedLineCoordIndexes(float pX, float pY) {
        List<Coordinate<Integer, Integer>> touchedPoints = new ArrayList<>();
        for (int i = 0; i < mTouchPoints.size(); i++) {
            for (int j = 0; j < mTouchPoints.get(i).size(); j++) {
                if (mTouchPoints.get(i).get(j).contains((int)pX, (int)pY))
                    touchedPoints.add(new Coordinate<>(i, j));
            }
        }

        return touchedPoints;
    }

    protected String getLineLabel(int pLineIndex) {
        return mLines.get(pLineIndex).getLabel();
    }

    protected Coordinate<Float, Float> getCoordinateAtIndexes(int pLineIndex, int pCoordinateIndex){
        return mLines.get(pLineIndex).getCoordinatesAtIndex(pCoordinateIndex);
    }

    protected void setChartBackgroundColor(int pColor) { mBackgroundColor = pColor; }

    protected void setGridlineWidth(float pGridlineWidth) { mGridlineWidth = pGridlineWidth; }

    protected void setNumOfGridlines(int pNumOfGridlines) { mNumOfGridlines = pNumOfGridlines; }

    protected void setGridlineColor(int pColor) { mGridlineColor = pColor; }

    protected void setRoundOutChartYMinMax (boolean pRound, int pHighestDigits) {
        mRoundOutChartYMinMax = pRound;
        mRoundOutChartYMinMaxHighestDigits =  pHighestDigits;
    }

    protected void setRoundOutChartXMinMax (boolean pRound, int pHighestDigits) {
        mRoundOutChartXMinMax = pRound;
        mRoundOutChartXMinMaxHighestDigits =  pHighestDigits;
    }

    protected void setZeroLineWidth(float pWidth) { mZeroLineWidth = pWidth; }

    protected void setZeroLineColor(int pColor) { mZeroLineColor = pColor; }

    protected void setBorderColor(int pColor) { mBorderColor = pColor; }

    protected void setBorderWidth(float pWidth) { mBorderWidth = pWidth; }

    protected void setUseAnimation(boolean pUseAnimation) { mUseAnimation = pUseAnimation; }

    protected void setAnimationTime(float pAnimationTime) { mAnimationTime = pAnimationTime;}

    protected int getLineColor(int pLineIndex) {
        return mLines.get(pLineIndex).getLineColor();
    }

    private void calculateScaling() {
        if (mRectF != null) {
            mRectF.set(mPadding, mPadding, getWidth() - mPadding, getHeight() - mPadding);

            mXScaling = (mRectF.width()) /(getXAxisLargestValue() - getXAxisSmallestValue());
            mYScaling = (mRectF.height())/(getYAxisLargestValue() - getYAxisSmallestValue());
        }
    }

    protected Float actualXToDrawingX(Float pActualX) {
        return mPadding+(mXScaling*(pActualX-getXAxisSmallestValue()));
    }

    protected Float actualYToDrawingY(Float pActualY) {
        return mRectF.height()+mPadding-(mYScaling*(pActualY-getYAxisSmallestValue()));
    }

    @SuppressWarnings("unused")
    protected Float drawingXToActualX(Float pDrawingX) {
        return ((pDrawingX-mPadding)/(mXScaling))+getXAxisSmallestValue();
    }

    @SuppressWarnings("unused")
    protected Float drawingYToActualY(Float pDrawingY) {
        return getYAxisSmallestValue()-((pDrawingY-mRectF.height()-mPadding)/mYScaling);
    }

    protected Float getXAxisSmallestValue() {
        if (mRoundOutChartXMinMax)
            return NumberUtilities.roundToNearestHighestPlaces(mMinimumXValue,
                    mRoundOutChartXMinMaxHighestDigits, false).floatValue();
        else
            return mMinimumXValue;
    }

    protected Float getXAxisLargestValue() {
        if (mRoundOutChartXMinMax)
            return NumberUtilities.roundToNearestHighestPlaces(mMaximumXValue,
                    mRoundOutChartXMinMaxHighestDigits, true).floatValue();
        else
            return mMaximumXValue;
    }

    protected Float getYAxisSmallestValue() {
        if (mRoundOutChartYMinMax)
            return NumberUtilities.roundToNearestHighestPlaces(mMinimumYValue,
                    mRoundOutChartYMinMaxHighestDigits, false).floatValue();
        else
            return mMinimumYValue;
    }

    protected Float getYAxisLargestValue() {
        if (mRoundOutChartYMinMax)
            return NumberUtilities.roundToNearestHighestPlaces(mMaximumYValue,
                    mRoundOutChartYMinMaxHighestDigits, true).floatValue();
        else
            return mMaximumYValue;
    }

    protected void reanimate() {
        mSweepX = 0f;
        mAnimationElapsedTime = 0;
        mAnimationStartTime = 0;
        mFinishedDrawing = false;
        mAnimationCounterStarted = false;
        invalidate();
    }

    private Float interpolateYFromX(Float pCurX, Coordinate<Float, Float> pCoord1,
                                                    Coordinate<Float, Float> pCoord2) {
        // get equation in form of y = mx+b and return what the y mValue is corresponding to pCurX
        Float m = (pCoord2.getY()-pCoord1.getY())/(pCoord2.getX()-pCoord1.getX());
        Float b = pCoord1.getY()-(m*pCoord1.getX());
        return m*pCurX+b;
    }
}
